"""Hold events."""
import logging

from camacq.const import BASE_EVENT

_LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class Event:
    """A base event.

    Parameters
    ----------
    data : dict
        The data of the event.

    Attributes
    ----------
    data : dict
        Return the data of the event.
    """

    __slots__ = ("data",)

    event_type = BASE_EVENT

    def __init__(self, data=None):
        """Set up event."""
        self.data = data or {}

    def __repr__(self):
        """Return the representation."""
        return f"{type(self).__name__}(data={self.data})"


class EventBus:
    """Representation of an eventbus.

    Parameters
    ----------
    center : Center instance
        The Center instance.
    """

    def __init__(self, center):
        """Set up instance."""
        self._center = center
        self._registry = {}

    @property
    def event_types(self):
        """:list: Return all registered event types."""
        return list(self._registry.keys())

    def _register_handler(self, event_type, handler):
        """Register handler to fire for events of type event_class."""
        handlers = self._registry.setdefault(event_type, [])
        handlers.append(handler)

    def register(self, event_type, handler):
        """Register event handler and return a function to remove it.

        An event can be a message from the microscope API or an
        internal event.

        Parameters
        ----------
        event_type : str
            A string representing the type of event.
        handler : callable
            A coroutine function that should accept two parameters, center and
            event. The first parameter is the Center instance, the
            second parameter is the Event instance that has fired.

        Returns
        -------
        callable
            Return a function to remove the registered handler.
        """
        _LOGGER.debug("Registering event handler for event type %s", event_type)
        self._register_handler(event_type, handler)

        def remove():
            """Remove registered event handler."""
            handlers = self._registry[event_type]
            try:
                handlers.remove(handler)
            except ValueError:
                _LOGGER.warning("Handler %s already removed from bus", handler)

        return remove

    async def notify(self, event):
        """Notify handlers that an event has fired.

        Parameters
        ----------
        event : Event instance
            An instance of Event or an instance of subclass of Event.
        """
        _LOGGER.debug("Notifying event %s", event)
        # Inspired by https://goo.gl/VEPG3n
        registry = self._registry
        for event_class in event.__class__.__mro__:
            # Handle base objects for Python 3.
            if event_class.__name__ == "object":
                continue
            for handler in registry.get(event_class.event_type, []):
                await handler(self._center, event)  # await in sequential order


def match_event(event, **event_data):
    """Return True if event attributes match event_data."""
    if not event_data or all(
        val == getattr(event, key, None) for key, val in event_data.items()
    ):
        return True

    return False



"""Host utils that are not aware of the implementation of camacq."""
import asyncio

try:
    asyncio_run = asyncio.run  # pylint: disable=invalid-name
except AttributeError:

    def asyncio_run(main, debug=False):
        """Mimic asyncio.run which is only in Python 3.7."""
        loop = asyncio.get_event_loop()
        loop.set_debug(debug)
        try:
            return loop.run_until_complete(main)
        finally:
            asyncio.set_event_loop(None)
            loop.close()


class dotdict(dict):  # pylint: disable=invalid-name
    """Access to dictionary attributes with dot notation."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



"""Control the microscope."""
import asyncio
import logging

from async_timeout import timeout as async_timeout
import voluptuous as vol

from camacq.event import Event, EventBus
from camacq.exceptions import CamAcqError, MissingActionError, MissingActionTypeError
from camacq.const import ACTION_TIMEOUT, CAMACQ_START_EVENT, CAMACQ_STOP_EVENT
from camacq.helper import register_signals
from camacq.plugins.sample import Samples
from camacq.util import dotdict

_LOGGER = logging.getLogger(__name__)


class Center:
    """Represent a control center for the microscope.

    Parameters
    ----------
    loop : asyncio.EventLoop
        The event loop.

    Attributes
    ----------
    loop : asyncio.EventLoop
        Return the event loop.
    bus : EventBus instance
        Return the EventBus instance.
    samples : Samples instance
        Return the Samples instance that holds all the Sample instances.
    actions : ActionsRegistry instance
        Return the ActionsRegistry instance.
    data : dict
        Return dict that stores data from other modules than control.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, loop=None):
        """Set up instance."""
        self.loop = loop or asyncio.get_event_loop()
        self.loop.set_exception_handler(loop_exception_handler)
        self.bus = EventBus(self)
        self.actions = ActionsRegistry(self)
        self.samples = Samples()
        self.data = {}
        self._exit_code = 0
        self._stopped = None
        self._pending_tasks = []
        self._track_tasks = False

    def __repr__(self):
        """Return the representation."""
        return "<Center>"

    async def end(self, code):
        """Prepare app for exit.

        Parameters
        ----------
        code : int
            Exit code to return when the app exits.
        """
        _LOGGER.info("Stopping camacq")
        self._track_tasks = True
        await self.bus.notify(CamAcqStopEvent({"exit_code": code}))
        self._exit_code = code
        await self.wait_for()
        if self._stopped is not None:
            self._stopped.set()
        else:
            self.loop.stop()

    async def start(self):
        """Start the app."""
        _LOGGER.info("Starting camacq")
        await self.bus.notify(CamAcqStartEvent())
        self._stopped = asyncio.Event()
        register_signals(self)
        await self._stopped.wait()
        return self._exit_code

    def add_executor_job(self, func, *args):
        """Schedule a function to be run in the thread pool.

        Return a task.
        """
        task = self.loop.run_in_executor(None, func, *args)

        if self._track_tasks:
            self._pending_tasks.append(task)

        return task

    def create_task(self, coro):
        """Schedule a coroutine on the event loop.

        Return a task.
        """
        task = self.loop.create_task(coro)

        if self._track_tasks:
            self._pending_tasks.append(task)

        return task

    async def wait_for(self):
        """Wait for all pending tasks."""
        await asyncio.sleep(0)
        while self._pending_tasks:
            _LOGGER.debug("Waiting for pending tasks")
            pending = [task for task in self._pending_tasks if not task.done()]
            self._pending_tasks.clear()
            if pending:
                await asyncio.wait(pending)
            else:
                await asyncio.sleep(0)


class ActionsRegistry:
    """Manage all registered actions."""

    def __init__(self, center):
        """Set up instance."""
        self._actions = {}
        self._center = center

    def __getattr__(self, action_type):
        """Return registered actions for an action type."""
        try:
            return self._actions[action_type]
        except KeyError as exc:
            raise MissingActionTypeError(action_type) from exc

    @property
    def actions(self):
        """:dict: Return dict of ActionTypes with all registered actions."""
        return self._actions

    def register(self, action_type, action_id, action_func, schema):
        """Register an action.

        Register actions per module.

        Parameters
        ----------
        action_type : str
            The name of the action_type to register the action under.
        action_id : str
            The id of the action to register.
        action_func : callable
            The function that should be called for the action.
        action_func : voluptuous schema
            The voluptuous schema that should validate the parameters
            of the action call.
        """
        if not asyncio.iscoroutinefunction(action_func):
            _LOGGER.error(
                "Action handler function %s is not a coroutine function", action_func
            )
            return
        if action_type not in self._actions:
            self._actions[action_type] = ActionType()
        _LOGGER.debug("Registering action %s.%s", action_type, action_id)
        self._actions[action_type][action_id] = Action(
            action_type, action_id, action_func, schema
        )

    async def call(self, action_type, action_id, **kwargs):
        """Call an action with optional kwargs.

        Parameters
        ----------
        action_type : str
            The name of the module where the action is registered.
        action_id : str
            The id of the action to call.
        **kwargs
            Arbitrary keyword arguments. These will be passed to the action
            function when an action is called.
        """
        if (
            action_type not in self._actions
            or action_id not in self._actions[action_type]
        ):
            _LOGGER.error(
                "No action registered for type %s or id %s", action_type, action_id
            )
            return

        action = self._actions[action_type][action_id]

        await action(**kwargs)


class ActionType(dotdict):
    """Represent an action type."""

    def __getattr__(self, action_id):
        """Return registered action for an action id."""
        try:
            return self[action_id]
        except KeyError as exc:
            raise MissingActionError(action_id) from exc


class Action:
    """Represent an action."""

    # pylint: disable=too-few-public-methods

    def __init__(self, action_type, action_id, func, schema):
        """Set up the instance."""
        self.action_type = action_type
        self.action_id = action_id
        self.func = func
        self.schema = schema

    async def __call__(self, **kwargs):
        """Call action."""
        silent = kwargs.get("silent", False)
        try:
            kwargs = self.schema(kwargs)
        except vol.Invalid as exc:
            _LOGGER.log(
                logging.DEBUG if silent else logging.ERROR,
                "Invalid action call parameters %s: %s for action: %s.%s",
                kwargs,
                exc,
                self.action_type,
                self.action_id,
            )
            return
        _LOGGER.log(
            logging.DEBUG if silent else logging.INFO,
            "Calling action %s.%s: %s",
            self.action_type,
            self.action_id,
            kwargs,
        )
        try:
            async with async_timeout(ACTION_TIMEOUT):
                await self.func(action_id=self.action_id, **kwargs)
        except asyncio.TimeoutError:
            _LOGGER.error(
                "Action %s.%s. timed out after %s seconds",
                self.action_type,
                self.action_id,
                ACTION_TIMEOUT,
            )
        except CamAcqError as exc:
            _LOGGER.error(
                "Failed to call action %s.%s: %s", self.action_type, self.action_id, exc
            )
            raise


def loop_exception_handler(loop, context):
    """Handle exceptions inside the event loop."""
    kwargs = {}
    exc = context.get("exception")
    if exc:
        kwargs["exc_info"] = exc

    _LOGGER.error("Error running job: %s", context["message"], **kwargs)


# pylint: disable=too-few-public-methods
class CamAcqStartEvent(Event):
    """An event fired when camacq has started."""

    __slots__ = ()

    event_type = CAMACQ_START_EVENT


class CamAcqStopEvent(Event):
    """An event fired when camacq is about to stop."""

    __slots__ = ()

    event_type = CAMACQ_STOP_EVENT

    @property
    def exit_code(self):
        """:int: Return the plate instance of the event."""
        return self.data.get("exit_code")



"""Handle logging."""
import logging
import logging.config
import logging.handlers
import os
from pathlib import Path

import colorlog

from camacq.const import CONFIG_DIR, LOG_LEVEL

LOG_FILENAME = "camacq.log"
_LOGGER = logging.getLogger(__name__)


def check_path(path):
    """Check that path to config exists and is writable for logging.

    Parameters
    ----------
    path : pathlib.Path
        The path to the log file or log directory.

    Returns
    -------
    bool
        Return True if path exists and is writable.
    """
    if (
        path.is_file()
        and os.access(path, os.W_OK)
        or path.parent.is_dir()
        and os.access(path.parent, os.W_OK)
    ):
        return True
    _LOGGER.error("Unable to access log file %s (access denied)", path)
    return False


def enable_log(config):
    """Enable logging.

    Parameters
    ----------
    config : dict
        The dict with the configuration.
    """
    logging.basicConfig(level=logging.INFO)
    root_logger = logging.getLogger()
    # basicConfig has added a StreamHandler
    # '%(log_color)s%(levelname)s:%(name)s:%(message)s'
    # '%(asctime)s;%(name)-16s;%(levelname)-8s;%(message)s'
    log_format = (
        "%(asctime)s;%(levelname)-5s;%(threadName)-10s;%(name)-18s;" "%(message)s"
    )
    date_format = "%Y-%m-%d %H:%M:%S"
    color_format = f"%(log_color)s{log_format}"
    root_logger.handlers[0].setFormatter(
        colorlog.ColoredFormatter(
            color_format,
            datefmt=date_format,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red",
            },
        )
    )
    log_config = config.get("logging")
    if log_config:
        log_path = log_config["handlers"]["filelog"]["filename"]
        if log_path and check_path(Path(log_path)):
            logging.config.dictConfig(log_config)
    else:  # get log path from default config dir
        _LOGGER.info("No config for logging supplied")
        log_path = config[CONFIG_DIR] / LOG_FILENAME
        _LOGGER.info("Using default log path at: %s", log_path)
        if check_path(log_path):
            filelog = logging.handlers.RotatingFileHandler(
                log_path, maxBytes=1048576, backupCount=9, encoding="utf-8", delay=0
            )
            filelog.setLevel(logging.WARNING)
            formatter = logging.Formatter(
                "%(asctime)s;%(name)-18s;%(levelname)-8s;%(message)s"
            )
            filelog.setFormatter(formatter)
            logging.getLogger("").addHandler(filelog)
    if LOG_LEVEL in config:
        log_level = config[LOG_LEVEL]
        root_logger.setLevel(log_level)
        root_logger.handlers[0].setLevel(config[LOG_LEVEL])
    # Silence matplotlib spam at debug level.
    logger = logging.getLogger("matplotlib")
    logger.setLevel(logging.INFO)



"""Leica microscope API specific modules."""
import asyncio
import logging
import tempfile
from functools import partial

import voluptuous as vol
from async_timeout import timeout as async_timeout
from leicacam.async_cam import AsyncCAM
from leicacam.cam import bytes_as_dict, check_messages, tuples_as_bytes
from leicaimage import attribute, attribute_as_str

from camacq.const import CAMACQ_STOP_EVENT
from camacq.helper import ensure_dict
from camacq.plugins.api import (
    Api,
    CommandEvent,
    ImageEvent,
    StartCommandEvent,
    StopCommandEvent,
    register_api,
)

from .command import start, stop
from .helper import find_image_path, get_field, get_imgs
from .sample import setup_module as sample_setup_module

_LOGGER = logging.getLogger(__name__)

CONF_HOST = "host"
CONF_IMAGING_DIR = "imaging_dir"
CONF_LEICA = "leica"
CONF_PORT = "port"
JOB_ID = "--E{:02d}"
LEICA_COMMAND_EVENT = "leica_command_event"
LEICA_START_COMMAND_EVENT = "leica_start_command_event"
LEICA_STOP_COMMAND_EVENT = "leica_stop_command_event"
LEICA_IMAGE_EVENT = "leica_image_event"
REL_IMAGE_PATH = "relpath"
SCAN_FINISHED = "scanfinished"
SCAN_STARTED = "scanstart"
START_STOP_DELAY = 2.0

CONFIG_SCHEMA = vol.Schema(
    vol.All(
        ensure_dict,
        {
            vol.Optional(CONF_HOST, default="localhost"): vol.Coerce(str),
            vol.Optional(CONF_PORT, default=8895): vol.Coerce(int),
            # pylint: disable=no-value-for-parameter
            vol.Optional(CONF_IMAGING_DIR, default=tempfile.gettempdir()): vol.IsDir(),
        },
    )
)


async def setup_module(center, config):
    """Set up Leica api package.

    Parameters
    ----------
    center : Center instance
        The Center instance.
    config : dict
        The config dict.
    """
    await sample_setup_module(center, config)
    conf = config[CONF_LEICA]
    host = conf[CONF_HOST]
    port = conf[CONF_PORT]
    cam = AsyncCAM(host, port, loop=center.loop)
    try:
        await cam.connect()
    except OSError as exc:
        _LOGGER.error("Connecting to server %s failed: %s", host, exc)
        return
    api = LeicaApi(center, conf, cam)
    register_api(center, api)
    # Start task that calls receive on the socket to the microscope
    task = center.create_task(api.start_listen())

    async def stop_listen(center, event):
        """Stop the task that listens to the client socket."""
        task.cancel()
        api.client.close()

    center.bus.register(CAMACQ_STOP_EVENT, stop_listen)


class LeicaApi(Api):
    """Represent the Leica API."""

    def __init__(self, center, config, client):
        """Set up the Leica API."""
        self.center = center
        self.client = client
        self.config = config
        self._last_image_path = None

    @property
    def name(self):
        """Return the name of the API."""
        return __name__

    async def start_listen(self):
        """Receive from the microscope socket."""
        try:
            while True:
                reply = await self.client.receive()
                self.center.create_task(self.receive(reply))
        except asyncio.CancelledError:
            _LOGGER.debug("Stopped listening for messages from CAM")

    async def receive(self, replies):
        """Receive replies from CAM server and fire an event per reply.

        Parameters
        ----------
        replies : list
            A list of replies from the CAM server.
        """
        # if reply check reply and call correct listener
        # parse reply and create Event
        # await event notify in sequential order
        # reply must be an iterable
        if not isinstance(replies, list):
            replies = [replies]
        for reply in replies:
            if not reply or not isinstance(reply, dict):
                continue
            if REL_IMAGE_PATH in reply:
                imaging_dir = self.config[CONF_IMAGING_DIR]
                rel_path = reply[REL_IMAGE_PATH]
                if rel_path == self._last_image_path:
                    # guard against duplicate image events from the microscope
                    _LOGGER.debug("Duplicate image reply received: %s", rel_path)
                    continue
                self._last_image_path = rel_path
                image_path = find_image_path(rel_path, imaging_dir)
                field_path = await self.center.add_executor_job(get_field, image_path)
                image_paths = await self.center.add_executor_job(
                    partial(
                        get_imgs,
                        field_path,
                        search=JOB_ID.format(attribute(image_path, "E")),
                    )
                )
                for path in image_paths:
                    # await in sequential order
                    await self.center.bus.notify(LeicaImageEvent({"path": path}))
            elif SCAN_STARTED in list(reply.values()):
                await self.center.bus.notify(LeicaStartCommandEvent(reply))
            elif SCAN_FINISHED in list(reply.values()):
                await self.center.bus.notify(LeicaStopCommandEvent(reply))
            else:
                await self.center.bus.notify(LeicaCommandEvent(reply))

    async def send(self, command, **kwargs):
        """Send a command to the Leica API.

        Parameters
        ----------
        command : list of tuples or string
            The command to send.
        """
        block = kwargs.get("block", True)

        if isinstance(command, str):
            command = bytes_as_dict(command.encode())
            command = list(command.items())
        cmd, value = command[0]  # use the first cmd and value to wait for
        cmd_sent = self.center.loop.create_future()

        async def receive_reply(center, event):
            """Indicate that reply has been received."""
            if check_messages([event.data], cmd, value=value):
                if not cmd_sent.done():
                    cmd_sent.set_result(True)

        remove = self.center.bus.register(LEICA_COMMAND_EVENT, receive_reply)
        cmd_sent.add_done_callback(lambda x: remove())

        await self.client.send(command)

        if not block:
            return cmd_sent
        return await cmd_sent

    async def start_imaging(self):
        """Send a command to the microscope to start the imaging."""
        await self._start_stop_imaging(start(), LEICA_START_COMMAND_EVENT, SCAN_STARTED)
        # A delay is needed after starting.
        await asyncio.sleep(START_STOP_DELAY)

    async def stop_imaging(self):
        """Send a command to the microscope to stop the imaging."""
        # A delay is needed before and after stopping.
        await asyncio.sleep(START_STOP_DELAY)
        await self._start_stop_imaging(stop(), LEICA_STOP_COMMAND_EVENT, SCAN_FINISHED)
        await asyncio.sleep(START_STOP_DELAY)

    async def _start_stop_imaging(self, cmd, event, ack_cmd):
        """Send a command to the microscope to start or stop the imaging."""
        cmd_sent = self.center.loop.create_future()

        async def receive_reply(center, event):
            """Indicate that reply has been received."""
            if not cmd_sent.done():
                cmd_sent.set_result(True)

        remove = self.center.bus.register(event, receive_reply)
        cmd_sent.add_done_callback(lambda x: remove())

        trigger_cmd_sent = await self.send(cmd, block=False)
        _LOGGER.info("Waiting for %s message for 10 seconds", ack_cmd)
        try:
            async with async_timeout(10.0):
                await asyncio.wait([cmd_sent, trigger_cmd_sent])
        except asyncio.TimeoutError:
            _LOGGER.info("No acknowledgement event received, continuing anyway")


# pylint: disable=too-few-public-methods
class LeicaCommandEvent(CommandEvent):
    """Leica CommandEvent class."""

    __slots__ = ()

    event_type = LEICA_COMMAND_EVENT

    @property
    def command(self):
        """Return the command string."""
        return tuples_as_bytes(list(self.data.items())).decode()


class LeicaStartCommandEvent(StartCommandEvent, LeicaCommandEvent):
    """Leica StartCommandEvent class."""

    __slots__ = ()

    event_type = LEICA_START_COMMAND_EVENT


class LeicaStopCommandEvent(StopCommandEvent, LeicaCommandEvent):
    """Leica StopCommandEvent class."""

    __slots__ = ()

    event_type = LEICA_STOP_COMMAND_EVENT


class LeicaImageEvent(ImageEvent):
    """Leica ImageEvent class."""

    __slots__ = ()

    event_type = LEICA_IMAGE_EVENT

    @property
    def path(self):
        """:str: Return absolute path to the image."""
        return self.data.get("path", "")

    @property
    def well_x(self):
        """:int: Return x coordinate of the well of the image."""
        return attribute(self.path, "U")

    @property
    def well_y(self):
        """:int: Return y coordinate of the well of the image."""
        return attribute(self.path, "V")

    @property
    def field_x(self):
        """:int: Return x coordinate of the well of the image."""
        return attribute(self.path, "X")

    @property
    def field_y(self):
        """:int: Return y coordinate of the well of the image."""
        return attribute(self.path, "Y")

    @property
    def z_slice_id(self):
        """:int: Return z index of the image."""
        return attribute(self.path, "Z")

    @property
    def channel_id(self):
        """:int: Return channel id of the image."""
        return attribute(self.path, "C")

    @property
    def job_id(self):
        """:int: Return job id of the image."""
        return attribute(self.path, "E")

    @property
    def plate_name(self):
        """:str: Return plate name of the image."""
        return attribute_as_str(self.path, "S")



"""Handle event trigger in automations."""
# Copyright 2013-2017 The Home Assistant Authors
# https://github.com/home-assistant/home-assistant/blob/master/LICENSE.md
# This file was modified by The Camacq Authors.
import logging

from camacq.const import CONF_ID, CONF_TRIGGER
from camacq.event import match_event

from . import CONF_TYPE

_LOGGER = logging.getLogger(__name__)

ATTR_EVENT = "event"
CONF_EVENT_DATA = "data"
CONF_EVENT = "event"


def handle_trigger(center, config, trigger_func):
    """Listen for events."""
    event_type = config[CONF_ID]
    event_data = config.get(CONF_EVENT_DATA, {})

    async def handle_event(center, event):
        """Listen for events and call trigger when data matches."""
        if match_event(event, **event_data):
            _LOGGER.debug("Trigger matched for event %s", event_type)
            # pass variables from trigger with event
            await trigger_func(
                {CONF_TRIGGER: {CONF_TYPE: CONF_EVENT, ATTR_EVENT: event}}
            )

    return center.bus.register(event_type, handle_event)



"""Provide exceptions."""


class CamAcqError(Exception):
    """Represent the base camacq exception."""


class MissingActionTypeError(CamAcqError):
    """Represent a missing action type error."""

    def __init__(self, action_type):
        """Set up the error."""
        super().__init__(f"No such action type registered: {action_type}")


class MissingActionError(CamAcqError):
    """Represent a missing action error."""

    def __init__(self, action_id):
        """Set up the error."""
        super().__init__(f"No such action id registered: {action_id}")


class SampleError(CamAcqError):
    """Represent a sample error."""


class TemplateError(CamAcqError):
    """Represent a template error."""

    def __init__(self, exc):
        """Set up the error."""
        super().__init__(f"{exc.__class__.__name__}: {exc}")



"""Handle commands."""


def start():
    """Start the scan.

    Return a list with parts for the cam command.
    """
    return [("cmd", "startscan")]


def stop():
    """Stop the scan.

    Return a list with parts for the cam command.
    """
    return [("cmd", "stopscan")]


def del_com():
    """Delete the cam list.

    Return a list with parts for the cam command.
    """
    return [("cmd", "deletelist")]


def camstart_com(afjob=None, afrange=None, afsteps=None):
    """Start the cam scan with selected AF job and AF settings.

    Return a list with parts for the cam command.
    """
    if afjob is not None:
        afjob = ("afj", str(afjob))
    if afrange is not None:
        afrange = ("afr", str(afrange))
    if afsteps is not None:
        afsteps = ("afs", str(afsteps))

    template = [("cmd", "startcamscan"), ("runtime", "36000"), ("repeattime", "36000")]

    for cmd in [afjob, afrange, afsteps]:
        if cmd:
            template.append(cmd)

    return template


def camstop_com():
    """Stop the cam scan.

    Return a list with parts for the cam command.
    """
    return [("cmd", "stopcamscan")]


def gain_com(exp, num, value):
    """Change the pmt gain in a job.

    Return a list with parts for the cam command.
    """
    return [
        ("cmd", "adjust"),
        ("tar", "pmt"),
        ("num", str(num)),
        ("exp", str(exp)),
        ("prop", "gain"),
        ("value", str(value)),
    ]


def enable_com(wellu, wellv, fieldx, fieldy, enable):
    """Enable a field in a well.

    Return a list with parts for the cam command.
    """
    wellx = str(wellu + 1)
    welly = str(wellv + 1)
    fieldx = str(fieldx + 1)
    fieldy = str(fieldy + 1)

    return [
        ("cmd", "enable"),
        ("slide", "0"),
        ("wellx", wellx),
        ("welly", welly),
        ("fieldx", fieldx),
        ("fieldy", fieldy),
        ("value", str(enable).lower()),
    ]


def cam_com(exp, wellu, wellv, fieldx, fieldy, dxcoord, dycoord):
    """Add a field to the cam list.

    Return a list with parts for the cam command.
    """
    # pylint: disable=too-many-arguments
    wellx = str(wellu + 1)
    welly = str(wellv + 1)
    fieldx = str(fieldx + 1)
    fieldy = str(fieldy + 1)

    return [
        ("cmd", "add"),
        ("tar", "camlist"),
        ("exp", exp),
        ("ext", "af"),
        ("slide", "0"),
        ("wellx", wellx),
        ("welly", welly),
        ("fieldx", fieldx),
        ("fieldy", fieldy),
        ("dxpos", str(dxcoord)),
        ("dypos", str(dycoord)),
    ]



"""Microscope API specific modules."""
import asyncio
import json
import logging

import voluptuous as vol

from camacq.const import (
    COMMAND_EVENT,
    IMAGE_EVENT,
    START_COMMAND_EVENT,
    STOP_COMMAND_EVENT,
)
from camacq.event import Event
from camacq.helper import BASE_ACTION_SCHEMA

_LOGGER = logging.getLogger(__name__)

COMMAND_VALIDATOR = vol.Any([(str, str)], vol.Coerce(str))


def validate_commands(value):
    """Validate a template string via JSON."""
    if isinstance(value, str):
        try:
            return json.loads(value)
        except ValueError:
            raise vol.Invalid(f"Invalid commands: {value}")
    else:
        schema = vol.Schema([COMMAND_VALIDATOR])
        return schema(value)


ACTION_SEND = "send"
ACTION_SEND_MANY = "send_many"
ACTION_START_IMAGING = "start_imaging"
ACTION_STOP_IMAGING = "stop_imaging"
CONF_API = "api"
DATA_API = "api"

SEND_ACTION_SCHEMA = BASE_ACTION_SCHEMA.extend(
    {"api_name": vol.Coerce(str), "command": COMMAND_VALIDATOR}
)

SEND_MANY_ACTION_SCHEMA = BASE_ACTION_SCHEMA.extend({"commands": validate_commands})

START_IMAGING_ACTION_SCHEMA = STOP_IMAGING_ACTION_SCHEMA = BASE_ACTION_SCHEMA

ACTION_TO_METHOD = {
    ACTION_SEND: {"method": "send", "schema": SEND_ACTION_SCHEMA},
    ACTION_SEND_MANY: {"method": "send_many", "schema": SEND_MANY_ACTION_SCHEMA},
    ACTION_START_IMAGING: {
        "method": "start_imaging",
        "schema": START_IMAGING_ACTION_SCHEMA,
    },
    ACTION_STOP_IMAGING: {
        "method": "stop_imaging",
        "schema": STOP_IMAGING_ACTION_SCHEMA,
    },
}


def register_api(center, api):
    """Register api."""
    api_store = center.data.setdefault(DATA_API, {})
    api_store[api.name] = api


async def setup_module(center, config):
    """Set up the microscope API package.

    Parameters
    ----------
    center : Center instance
        The Center instance.
    config : dict
        The config dict.
    """
    api_store = center.data.setdefault(DATA_API, {})

    async def handle_action(**kwargs):
        """Handle action call to send a command to an api.

        Parameters
        ----------
        **kwargs
            Arbitrary keyword arguments. These will be passed to the
            api method when an action is called.
        """
        action_id = kwargs.pop("action_id")
        method = ACTION_TO_METHOD[action_id]["method"]
        api_name = kwargs.pop("api_name", None)
        if api_name:
            apis = [api_store[api_name]]
        else:
            apis = list(api_store.values())
        tasks = []
        for api in apis:
            _LOGGER.debug("Handle API %s action %s: %s", api.name, action_id, kwargs)
            tasks.append(center.create_task(getattr(api, method)(**kwargs)))
        if tasks:
            await asyncio.wait(tasks)

    for action_id, options in ACTION_TO_METHOD.items():
        schema = options["schema"]
        center.actions.register("command", action_id, handle_action, schema)


class Api:
    """Represent the microscope API."""

    @property
    def name(self):
        """Return the name of the API."""
        raise NotImplementedError()

    async def send(self, command, **kwargs):
        """Send a command to the microscope API.

        Parameters
        ----------
        command : str
            The command to send.
        """
        raise NotImplementedError()

    async def send_many(self, commands, **kwargs):
        """Send multiple commands to the microscope API.

        Parameters
        ----------
        commands : list
            A list of commands to send.
        """
        for cmd in commands:
            # It's important that each task is done before we start the next.
            await self.send(cmd, **kwargs)

    async def start_imaging(self):
        """Send a command to the microscope to start the imaging."""
        raise NotImplementedError()

    async def stop_imaging(self):
        """Send a command to the microscope to stop the imaging."""
        raise NotImplementedError()


# pylint: disable=too-few-public-methods
class CommandEvent(Event):
    """An event received from the API.

    Notify with this event when a command is received via API.
    """

    __slots__ = ()

    event_type = COMMAND_EVENT

    @property
    def command(self):
        """:str: Return the command string."""
        return None


class StartCommandEvent(CommandEvent):
    """An event received from the API.

    Notify with this event when imaging starts via API.
    """

    __slots__ = ()

    event_type = START_COMMAND_EVENT


class StopCommandEvent(CommandEvent):
    """An event received from the API.

    Notify with this event when imaging stops via API.
    """

    __slots__ = ()

    event_type = STOP_COMMAND_EVENT


class ImageEvent(Event):
    """An event received from the API.

    Notify with this event when an image is saved via API.
    """

    __slots__ = ()

    event_type = IMAGE_EVENT

    @property
    def path(self):
        """:str: Return absolute path to the image."""
        return self.data.get("path")

    @property
    def well_x(self):
        """:int: Return x coordinate of the well of the image."""
        return self.data.get("well_x")

    @property
    def well_y(self):
        """:int: Return y coordinate of the well of the image."""
        return self.data.get("well_y")

    @property
    def field_x(self):
        """:int: Return x coordinate of the well of the image."""
        return self.data.get("field_x")

    @property
    def field_y(self):
        """:int: Return y coordinate of the well of the image."""
        return self.data.get("field_y")

    @property
    def z_slice_id(self):
        """:int: Return z index of the image."""
        return self.data.get("z_slice_id")

    @property
    def channel_id(self):
        """:int: Return channel id of the image."""
        return self.data.get("channel_id")

    @property
    def plate_name(self):
        """:str: Return plate name of the image."""
        return self.data.get("plate_name")

    def __repr__(self):
        """Return the representation."""
        data = dict(
            plate_name=self.plate_name,
            well_x=self.well_x,
            well_y=self.well_y,
            field_x=self.field_x,
            field_y=self.field_y,
            z_slice_id=self.z_slice_id,
            channel_id=self.channel_id,
        )
        return f"{type(self).__name__}(data={data})"



"""Helper functions for camacq."""
import logging
import pkgutil
import signal
import sys
from importlib import import_module

import voluptuous as vol

import camacq

_LOGGER = logging.getLogger(__name__)

PACKAGE_MODULE = "{}.{}"
BASE_ACTION_SCHEMA = vol.Schema(
    {
        "action_id": str,
        "silent": vol.Boolean(),  # pylint: disable=no-value-for-parameter
    },
    extra=vol.REMOVE_EXTRA,
)


def get_module(package, module_name):
    """Return a module from a package.

    Parameters
    ----------
    package : str
        The path to the package.
    module_name : str
        The name of the module.
    """
    module_path = PACKAGE_MODULE.format(package, module_name)
    matches = [
        name
        for _, name, _ in pkgutil.walk_packages(
            camacq.__path__, prefix=f"{camacq.__name__}."
        )
        if module_path in name
    ]
    if len(matches) != 1:
        raise ValueError("Invalid module search result, not a single match")
    module_path = matches[0]
    try:
        module = import_module(module_path)
        _LOGGER.debug("Loaded %s from %s", module_name, module_path)

        return module

    except ImportError:
        _LOGGER.exception(("Loading %s failed"), module_path)


async def setup_one_module(center, config, module):
    """Set up one module or package.

    Returns
    -------
    asyncio.Task
        Return a task to set up the module or None.
    """
    module_name = module.__name__.split(".")[-1]
    if not hasattr(module, "setup_module"):
        _LOGGER.warning("Missing setup_module function in module %s", module_name)
        return
    _LOGGER.info("Setting up module %s", module_name)
    if hasattr(module, "CONFIG_SCHEMA"):
        _LOGGER.debug("Validating config for module %s", module_name)
        module_conf = config[module_name]
        try:
            module_conf = await center.add_executor_job(
                module.CONFIG_SCHEMA, module_conf
            )
        except vol.Invalid:
            _LOGGER.exception("Incorrect configuration for module %s:", module_name)
            return
        config[module_name] = module_conf
    await module.setup_module(center, config)


# Adapted from:
# https://github.com/alecthomas/voluptuous/issues/115#issuecomment-144464666
def has_at_least_one_key(*keys):
    """Validate that at least one key exists."""

    def validate(obj):
        """Test keys exist in dict."""
        if not isinstance(obj, dict):
            raise vol.Invalid("expected a dictionary")

        for key in obj:
            if key in keys:
                return obj
        all_keys = ", ".join(keys)
        raise vol.Invalid(f"must contain at least one of {all_keys}")

    return validate


def ensure_dict(value):
    """Convert None to empty dict."""
    if value is None:
        return {}
    return value


def register_signals(center):
    """Register signal handlers."""
    if sys.platform != "win32":

        def handle_signal(exit_code):
            """Handle a signal."""
            center.loop.remove_signal_handler(signal.SIGTERM)
            center.loop.remove_signal_handler(signal.SIGINT)
            center.create_task(center.end(exit_code))

        center.loop.add_signal_handler(signal.SIGTERM, handle_signal, 0)
        center.loop.add_signal_handler(signal.SIGINT, handle_signal, 0)

    else:

        prev_sig_term = None
        prev_sig_int = None

        def handle_signal(signum, frame):
            """Handle a signal."""
            signal.signal(signal.SIGTERM, prev_sig_term)
            signal.signal(signal.SIGINT, prev_sig_int)
            center.create_task(center.end(signum))

        prev_sig_term = signal.signal(signal.SIGTERM, handle_signal)
        prev_sig_int = signal.signal(signal.SIGINT, handle_signal)



"""Handle the config file."""
import logging
import os
from pathlib import Path

from pkg_resources import resource_filename
from ruamel.yaml import YAML, YAMLError

_LOGGER = logging.getLogger(__name__)

CONFIG_DIR_NAME = ".camacq"
YAML_CONFIG_FILE = "config.yml"
DEFAULT_CONFIG_TEMPLATE = "data/config.yml"


def get_default_config_dir():
    """Get the default configuration directory based on OS.

    Returns
    -------
    str
        Return the path to the configuration directory.
    """
    data_dir = os.getenv("APPDATA") if os.name == "nt" else Path.home()
    return (Path(data_dir) / CONFIG_DIR_NAME).resolve()


def find_config_file(config_dir):
    """Find the configuration file in the configuration directory.

    Parameters
    ----------
    config_dir : pathlib.Path
        The path to the configuration directory.

    Returns
    -------
    pathlib.Path
        Return path to the configuration file if found, None if not
        found.
    """
    config_path = config_dir / YAML_CONFIG_FILE

    return config_path if config_path.is_file() else None


def load_config_file(path):
    """Parse a YAML configuration file.

    Parameters
    ----------
    path : pathlib.Path
        The path to the configuration YAML file.

    Returns
    -------
    dict
        Return a dict with the configuration contents.
    """
    yaml = YAML()
    try:
        with open(path, "r") as yml_file:
            cfg = yaml.load(yml_file)
    except YAMLError as exception:
        _LOGGER.error("Error reading YAML configuration file %s", path)
        raise YAMLError(exception)  # or let it pass?
    if not isinstance(cfg, dict):
        _LOGGER.error(
            "The configuration file %s does not contain a dictionary", path.name,
        )
        raise TypeError()  # or let it pass?
    return cfg


def create_default_config(config_dir):
    """Create a default config file in given configuration directory.

    Parameters
    ----------
    config_dir : pathlib.Path
        The path to the configuration directory.

    Returns
    -------
    pathlib.Path
        Return path to new configuration file if success, None if
        failed.
    """
    config_path = config_dir / YAML_CONFIG_FILE
    default_config_template = resource_filename(__name__, DEFAULT_CONFIG_TEMPLATE)
    data = load_config_file(Path(default_config_template))
    yaml = YAML()

    try:
        with open(config_path, "w") as config_file:
            yaml.dump(data, config_file)
    except OSError:
        _LOGGER.error("Unable to create default configuration file %s", config_path)
        return None

    return config_path


def ensure_config_exists(config_dir):
    """Ensure configuration file exists in the configuration directory.

    Create a default configuration file if needed.

    Parameters
    ----------
    config_dir : pathlib.Path
        The path to the configuration directory.

    Returns
    -------
    pathlib.Path
        Return path to the configuration file.
    """
    config_path = find_config_file(config_dir)

    if config_path is None:
        print("Unable to find configuration. Creating default one in", config_dir)
        config_path = create_default_config(config_dir)

    return config_path



"""Helper functions for Leica api."""
from pathlib import Path, PureWindowsPath

from leicaimage import experiment


def find_image_path(relpath, root):
    """Parse the relpath from the server to find file path from root.

    Convert from windows path to posix path.

    Parameters
    ----------
    relpath : str
        A relative path to the image.
    root : str
        Path to directory where path should start.

    Returns
    -------
    str
        Return path to image.
    """
    parts = PureWindowsPath(relpath).parts
    return str(Path(root).joinpath(*parts))


def get_field(path):
    """Get path to field from image path.

    Parameters
    ----------
    path : string
        Path to image.

    Returns
    -------
    str
        Return path to field directory of image.
    """
    return experiment.Experiment(path).dirname  # pylint: disable=no-member


def get_well(path):
    """Get path to well from image path.

    Parameters
    ----------
    path : string
        Path to image.

    Returns
    -------
    str
        Return path to well directory of image.
    """
    # pylint: disable=no-member
    return experiment.Experiment(get_field(path)).dirname


def get_imgs(path, img_type="tif", search=""):
    """Get all images below path.

    Parameters
    ----------
    path : string
        Path to directory where to search for images.
    img_type : string
        A string representing the image file type extension.
    path : string
        A glob pattern string to use in the search.

    Returns
    -------
    list
        Return paths of all images found.
    """
    root = Path(path)
    _path = Path("")
    if search:
        search = f"{search}*"
    patterns = ["slide", "chamber--", "field--", "image--"]
    for pattern in patterns:
        if pattern not in path:
            _path = _path / f"{pattern}*"
    return list(root.glob(f"{_path}{search}.{img_type}"))
